import bs
import bsUtils
import bsTeamGame
import bsMap
import bsVector
import bsFlag
import bsScoreBoard
import bsSpaz
import random
import weakref


def getGames():
    return [ConquestGame]


class ConquestFlag(bsFlag.Flag):
    def __init__(self,*args,**keywds):
        bsFlag.Flag.__init__(self,*args,**keywds)
        self._team = None
    def setTeam(self,team):
        self._team = None if team is None else weakref.ref(team)
    def getTeam(self):
        return None if self._team is None else self._team()


class ConquestGame(bsTeamGame.TeamGameActivity):

    @classmethod
    def getName(c):
        return 'Conquest'

    @classmethod
    def getDescription(c,sessionType):
        return 'Secure all flags on the map to win.'

    @classmethod
    def getSupportedSessionTypes(self):
        return ['teams']

    @classmethod
    def getSupportedMaps(self,sessionType):
        return bsMap.getMapList("conquest")

    @classmethod
    def getSettings(self,sessionType):
        return [("Time Limit",{'choices':[('None',0),('1 Minute',60),
                                        ('2 Minutes',120),('5 Minutes',300),
                                        ('10 Minutes',600),('20 Minutes',1200)],'default':0}),
                ("Respawn Times",{'choices':[('Shorter',0.25),('Short',0.5),('Normal',1.0),('Long',2.0),('Longer',4.0)],'default':1.0}),
                ("Epic Mode",{'default':False})]

    def __init__(self,session,level,settings):
        bsTeamGame.TeamGameActivity.__init__(self,session,level,settings)
        if self.settings['Epic Mode']: self._isSlowMotion = True
        self._scoreBoard = bsScoreBoard.ScoreBoard(self)
        self._scoreSound = bs.getSound("score")
        self._swipSound = bs.getSound("swip")

        self._extraFlagMaterial = bs.Material('Conquest Flag')

        # we want flags to tell us they've been hit but not react physically
        self._extraFlagMaterial.addAction(conditions=("dstIsMaterial",bs.playerMaterial),
                                          actions=(("partMod","collide",True),
                                                   ("call","atConnect",bs.WeakCall(self._handleFlagPlayerCollide))))

    # def getInstanceName(self):
    #     return ("Epic " if self.settings['Epic Mode'] else "") + self.getName()

    def getInstanceDescription(self):
        return ('Secure all ${ARG1} flags.',len(self._level.flagPoints))
        # epicStr = ' In epic slow motion.' if self.settings['Epic Mode'] else ''
        # return 'Secure all '+str(len(self._level.flagPoints))+' flags.'+epicStr

    def getInstanceScoreBoardDescription(self):
        return ('secure all ${ARG1} flags',len(self._level.flagPoints))
        #return 'secure all '+str(len(self._level.flagPoints))+' flags'

    def onTransitionIn(self):
        bsTeamGame.TeamGameActivity.onTransitionIn(self, music='Epic' if self.settings['Epic Mode'] else 'GrandRomp')

    def onTeamJoined(self,team):
        if self.hasBegun(): self._updateScores()
        team.data['flagsHeld'] = 0
        
    def onPlayerJoined(self,player):
        player.data['respawnTimer'] = None
        # only spawn if this player's team has a flag currently
        if player.getTeam().data['flagsHeld'] > 0: self.spawnPlayer(player)

    def onBegin(self):
        bsTeamGame.TeamGameActivity.onBegin(self,nameSound=bs.getSound('announceConquest'))

        self.setupStandardTimeLimit(self.settings['Time Limit'])

        # set up flags with marker lights..
        self._flags = []
        for i in range(len(self._level.flagPoints)):
            point = self._level.flagPoints[i]
            flag = ConquestFlag(self,bs.Actor.UNRETAINED,
                                translate=point,touchable=False,
                                materials=(self._extraFlagMaterial.materialID,))
            self.projectFlagStand(point)
            self._flags.append(flag)
            flag.id = i
            flag.light = bs.newNode("light",
                                    owner=flag.node,
                                    attrs={"translate": point,
                                           "intensity":0.25,
                                           "heightAttenuated":False,
                                           "radius":0.3,
                                           "color": (1,1,1)})

        # give teams a flag to start with
        for i in range(len(self.teams)):
            self._flags[i].setTeam(self.teams[i])
            self._flags[i].light.color = self.teams[i].color
            self._flags[i].node.color = self.teams[i].color

        self._updateScores()

        # initial joiners didn't spawn due to no flags being owned yet; spawn them now
        for player in self.players.values():
            self.spawnPlayer(player)

    def _updateScores(self):

        for team in self.teams.values(): team.data['flagsHeld'] = 0

        for flag in self._flags:
            try: flag.getTeam().data['flagsHeld'] += 1
            except Exception: pass

        for team in self.teams.values():
            # if a team finds themselves with no flags, cancel all outstanding spawn-timers
            if team.data['flagsHeld'] == 0:
                for player in team.players.values():
                    player.data['respawnTimer'] = None
                    player.data['respawnIcon'] = None
            if team.data['flagsHeld'] == len(self._flags):
                self.endGame()
            self._scoreBoard.setTeamValue(team,team.data['flagsHeld'],len(self._flags))

    def endGame(self):
        results = bsTeamGame.Results(scoreName='Flags')
        for t in self.teams.values(): results.setTeamScore(t,t.data['flagsHeld'])
        self.end(results=results)
        
    def _flashFlag(self,flag,length=1000):
        light = bs.newNode("light",owner=self,attrs={"translate": flag.node.translate,"heightAttenuated":False,"color": flag.light.color})
        bsUtils.animate(light,"intensity",{0:0,250:1,500:0},loop=True)
        bs.gameTimer(length,bs.WeakCall(bs.deleteNode,light))

    def _handleFlagPlayerCollide(self):
        flagNode,playerNode = bs.getCollideValues("sourceNode","opposingNode")
        try:
            player = playerNode.getDelegate().getPlayer()
            flag = flagNode.getDelegate()
        except Exception: return # player may have left and his body hit the flag

        if flag.getTeam() is not player.getTeam():
            flag.setTeam(player.getTeam())
            flag.light.color = player.getTeam().color
            flag.node.color = player.getTeam().color
            self.scoreSet.playerScored(player,10,screenMessage=False)
            bs.playSound(self._swipSound)
            self._flashFlag(flag)
            self._updateScores()

            # respawn any players on this team that were in limbo due to the lack of a flag for their team
            for p in self.players.values():
                if p.getTeam() is flag.getTeam() and p.actor is not None and p.actor.dead and p.data['respawnTimer'] is None:
                    self.spawnPlayer(p)

    def handleMessage(self,m):
        if isinstance(m,bsSpaz.PlayerSpazDeathMessage):
            self.__superHandleMessage(m) # augment standard behavior

            player = m.spaz.getPlayer()

            # respawn only if this team has a flag
            if player is not None and player.getTeam().data['flagsHeld'] > 0: self.respawnPlayer(m.spaz.getPlayer())
            else: player.data['respawnTimer'] = None

        else: self.__superHandleMessage(m)

    def __superHandleMessage(self,m):
        super(ConquestGame,self).handleMessage(m)

    def spawnPlayer(self,player):
        # we spawn players at different places based on what flags are held
        self.spawnSpazForPlayer(player,self._getPlayerSpawnPosition(player))

    def _getPlayerSpawnPosition(self,player):

        # iterate until we find a spawn owned by this team..
        spawnCount = len(self._level.spawnByFlagPoints)

        # get all spawns owned by this team
        spawns = [i for i in range(spawnCount) if self._flags[i].getTeam() is player.getTeam()]

        closestSpawn = 0
        closestDistance = 9999

        # now find the spawn thats closest to a spawn not owned by us.. we'll use that one
        for s in spawns:
            p = self._level.spawnByFlagPoints[s]
            ourPt = bsVector.Vector(p[0],p[1],p[2])
            for os in [i for i in range(spawnCount) if self._flags[i].getTeam() is not player.getTeam()]:
                p = self._level.spawnByFlagPoints[os]
                theirPt = bsVector.Vector(p[0],p[1],p[2])
                dist = (theirPt-ourPt).length()
                if dist < closestDistance:
                    closestDistance = dist
                    closestSpawn = s

        pt = self._level.spawnByFlagPoints[closestSpawn]
        xRange = (-0.5,0.5) if pt[3] == 0 else (-pt[3],pt[3])
        zRange = (-0.5,0.5) if pt[5] == 0 else (-pt[5],pt[5])
        pt = (pt[0]+random.uniform(*xRange),pt[1],pt[2]+random.uniform(*zRange))
        return pt
